//////////////////////////////////////////////////////////////////////////////////////
//  
//  FMPluginSolidColor.cpp - Sample external function for FileMaker plug-in
//
//  Copyright 2003-2005 by 24U Software. All rights reserved.
//
//  Written by Tomas Zahradnicky, HOnza Koudelka, Tomas Rycl
//
////////////////////////////////////////////////////////////////////////////////
//
//  This example is based on 24U FM Template, a stationery for building
//  FileMaker plug-ins from 24u Software.
//
//  The latest version of 24U FM Template is available for download from:
//
//  http://24uSoftware.com/FMTemplate  or  http://www.NewMillennium.com/
//
////////////////////////////////////////////////////////////////////////////////
//
//  24U Sample Code - Terms of use
//
//  You may incorporate this sample code into your products without
//  restriction, though the code has been provided "AS IS" and the
//  responsibility for its operation is 100% yours.
//
//  You are not however permitted to redistribute the source
//  as "24U Sample Code" after having made changes.
//
//  If you are going to redistribute the source with references to
//  24U Software, we require that you make it clear in the source that
//  the code was descended from 24U Sample Code, but that you've made changes.
//
////////////////////////////////////////////////////////////////////////

//
//  This function generates a combination of 1x1 JPEG and 1x1 PICT picture
//  or a combination of 1x1 JPEG and 255x255 PNG if optional parameter
//  opacity is defined.
//  All the pixels of generated pictures have a color specified by the red,
//  green and blue values. 
//
////////////////////////////////////////////////////////////////////////////////


#include "FMPlugin.h"
#include "FMTemplate/FMTemplate.h"
#include "PngCrc/pngcrc.h"

static FMX_Hdl CreateSinglePixelGIF( unsigned char R, unsigned char G, unsigned char B );
static FMX_Hdl CreateSinglePixelPICT( unsigned char R, unsigned char G, unsigned char B );
static FMX_Hdl CreateSinglePixelJPEG();
static FMX_Hdl CreateAlphaPNG( unsigned char R, unsigned char G, unsigned char B, unsigned char A );

#pragma mark Plugin_SolidColor
FMX_PROC(fmx::errcode) Plugin_SolidColor(short funcId, const fmx::ExprEnv& environment, const fmx::DataVect& dataVect, fmx::Data& result)
{
    #pragma unused(funcId,environment)                              // This function is supposed to work anywhere
    
    fmx::errcode            errorResult = -1;                       // Prepare an error for the case we fail
    unsigned long           colorRGB;                               // RGB color.
    unsigned char           red, green, blue;                       // Red, green and blue component of RGB
    fmx::BinaryDataAutoPtr  picture;                                // Prepare a picture stream
    fmx::QuadCharAutoPtr    sizeTypeID('S','I','Z','E');            // Type "SIZE"
    fmx::QuadCharAutoPtr    jpegTypeID('J','P','E','G');            // Type "JPEG"
    fmx::QuadCharAutoPtr    pictTypeID('P','I','C','T');            // Type "PICT"
    fmx::QuadCharAutoPtr    currTypeID('P','N','G','f');            // Current type, pretending PNG
    FMX_Hdl                 image   = NULL;                         // Various data types handlers
    FMX_Hdl                 jpeg    = NULL;
    unsigned char           PSize[4]= {0x00, 0xFF, 0x00, 0xFF};     // Current image size, pretending 255 x 255 pixel
    
    if (dataVect.Size() > 0) 
    {
        colorRGB = dataVect.AtAsNumber(0).AsLong();                 // Reading a color information from function parameters
        red   = (unsigned char)(colorRGB >> 16);
        green = (unsigned char)(colorRGB >> 8);
        blue  = (unsigned char) colorRGB;
        
        jpeg = CreateSinglePixelJPEG();                             // Get the data for the stream as JPEG
    
        if(jpeg)                                                    // Were we successfull getting the JPEG?
        {
            FMX_LockHandle(jpeg);                                   // Lock the JPEG's memory so that we can manipulate it
            
            picture -> Add(*jpegTypeID, 
                (fmx::ulong)FMX_GetHandleSize(jpeg), *jpeg);        // Add the JPEG to the picture stream
            
            FMX_UnlockHandle(jpeg);                                 // Unlock the JPEG's memory and dispose of it
            FMX_DisposeHandle(jpeg);                                //  as we don't need it any more
         
            switch ( dataVect.Size() )
            {
                case 1:
                default:
                    image = 
                        CreateSinglePixelPICT(red, green, blue);    // Get the data for the stream as PICT and store it into image
                    currTypeID = pictTypeID;                        // Change type to PICT
                    PSize[0] = 0x00; PSize[1] = 0x01;
                    PSize[2] = 0x00; PSize[3] = 0x01;               // Change image size from 255x255 to 1x1
                    break;
                case 2:
                    unsigned char opacity  = (unsigned char)
                        dataVect.AtAsNumber(1).AsLong();            // Reading an opacity information
                    image = 
                        CreateAlphaPNG(red, green, blue, opacity);  // Get the data for the stream as PNG and store it into image
            }
            
            if( image )                                             // Were we successfull getting the image?
            {
                FMX_LockHandle(image);                              // Lock the image's memory so that we can manipulate it
                
                picture -> Add(*currTypeID,
                    (fmx::ulong)FMX_GetHandleSize(image), *image);  // Add the image to the picture stream
                picture -> Add(*sizeTypeID, 4, &PSize);             // Add the size info to the stream
                result.SetBinaryData(*picture);                     // Set the function result to our picture stream

                FMX_UnlockHandle(image);                            // Unlock the images's memory and dispose of it,
                FMX_DisposeHandle(image);                           //  as we don't need it any more
                    
                errorResult = 0;                                    // We have successfully set the result => no error
            }
            else errorResult = FMX_MemoryError();                   // Image could not be created => find out why
        }
        else errorResult = FMX_MemoryError();                       // JPEG could not be created => find out why
    }
    return(errorResult);
}

static FMX_Hdl CreateSinglePixelGIF( unsigned char R, unsigned char G, unsigned char B )
{
    unsigned char onePixelGif[] = 
    { 
        0x47, 0x49, 0x46, 0x38, 0x37, 0x61,                         // Header (GIF87a)
        0x01, 0x00,                                                 // Width in little endian 
        0x01, 0x00,                                                 // Height in little endian
        0x80,                                                       // Flags 
        0x00,                                                       // Back color index
        0x00,                                                       // Filler
        0xFF, 0x00, 0x00,                                           // Red (RR GG BB)
        0xFF, 0xFF, 0xFF,                                           // White (RR GG BB)
        0x2C,                                                       // Begin marker
        0x00, 0x00,                                                 // Left in little endian
        0x00, 0x00,                                                 // Top in little endian
        0x01, 0x00,                                                 // Width in little endian
        0x01, 0x00,                                                 // Height in little endian
        0x00,                                                       // Filler
        0x02, 0x02, 0x44, 0x01,                                     // LZW
        0x00,                                                       // LZW end mark
        0x3B                                                        // End marker
    };
    
    FMX_Hdl result = FMX_NewHandle( sizeof(onePixelGif) );          // Allocate memory for one-pixel GIF
    
    if( result )
    {
        memcpy( *result, onePixelGif, sizeof(onePixelGif) );        // Copy default one-pixel GIF to the buffer
        
        (*result)[13] = (char)R;                                    // Modify the color of the one-pixel
        (*result)[14] = (char)G;
        (*result)[15] = (char)B;
    }
    
    return result;
}

static FMX_Hdl CreateSinglePixelPICT( unsigned char R, unsigned char G, unsigned char B )
{
    //Indexed color PICT. Width=1, Heigth=1
    //Contains only 1 true color pixel 
    
    unsigned char truePCT[] = 
    {
        0x00, 0x9C,                                                 // Size of file in bytes
        0x00, 0x00,                                                 // Top-left x of frame 
        0x00, 0x00,                                                 // Top-left y of frame
        0x00, 0x01,                                                 // Low-right x of frame
        0x00, 0x01,                                                 // Low-right y of frame
        0x00, 0x11,                                                 // Version code: version 2
        0x02, 0xFF,                                                 // Version number: version 2
        0x0C, 0x00,                                                 // Version header: 0C00=current
        0xFF, 0xFE,                                                 // Variable resolution
        0x00, 0x00,                                                 // Reserved
        0x00, 0x47, 0xFF, 0xB4,                                     // Original x resolution
        0x00, 0x47, 0xFF, 0xB4,                                     // Original y resolution
        0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 
        0x00, 0x01,                                                 // ??
        0x00, 0x00, 0x00, 0x00, 0x00, 
        0xA1, 0x01, 0xF2, 0x00, 0x16, 0x38, 0x42, 
        0x49, 0x4D, 0x00, 0x00, 0x00, 0x00, 0x00, 
        0x00, 0x00, 0x01, 0x00, 0x01, 0x47, 0x72, 
        0x89, 0x70, 0x68, 0x81, 0x81, 0xAF, 0x00, 
        0x01, 0x00, 0x0A, 0x00, 0x00, 0x00, 0x00, 
        0x00, 0x01, 0x00, 0x01, 0x00, 0x9A, 0x00, 
        0x00, 0x00, 0xFF, 0x80, 0x04, 0x00, 0x00, 
        0x00, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 
        0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 
        0x00, 0x47, 0xFF, 0xB4, 0x00, 0x47, 0xFF, 
        0xB4, 0x00, 0x10, 0x00, 0x20, 0x00, 0x03, 
        0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
        0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 
        0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 
        0x00, 0x01, 0x00, 0x40, 
        0x00,                                                       // Alpha channel (not in standard) 
        0xFF,                                                       // Red
        0x00,                                                       // Green
        0x00,                                                       // Blue
        0x00, 0xFF
    };
    
    FMX_Hdl result = FMX_NewHandle( sizeof(truePCT) );
 
    if( result )
    {
        memcpy( *result, truePCT, sizeof(truePCT) );
        
        (*result)[0x97] = (unsigned char)R;
        (*result)[0x98] = (unsigned char)G;
        (*result)[0x99] = (unsigned char)B;
    }
    
    return result;
}

static FMX_Hdl CreateAlphaPNG( unsigned char R, unsigned char G, unsigned char B, unsigned char A )
{
    //Indexed color PNG. Width=255, Heigth=255, with indexed alpha channel
    
    unsigned char solidPng[] = 
    {
        0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A,             // PNG signature
        0x00, 0x00, 0x00, 0x0D,                                     // IHDR chunk length 
        0x49, 0x48, 0x44, 0x52,                                     // IHDR, begin of header chunk
        0x00, 0x00, 0x00, 0xFF,                                     // Width, in pixels, big endian
        0x00, 0x00, 0x00, 0xFF,                                     // Heigth, in pixels
        0x01,                                                       // Bit depth
        0x03,                                                       // Color type: 03 - indexed colors from palette
        0x00,                                                       // Compression method
        0x00,                                                       // Filter method
        0x00,                                                       // Interlace method 
        0x04, 0xC6, 0x92, 0x89,                                     // IHDR CRC
        0x00, 0x00, 0x00, 0x03,                                     // PLTE chunk length
        0x50, 0x4C, 0x54, 0x45,                                     // PLTE, palette chunk beginning
        0x00, 0x00, 0xFF,                                           // Color 0 palette
        0x8A, 0x78, 0xD2, 0x57,                                     // PLTE CRC
        0x00, 0x00, 0x00, 0x01,                                     // tRNS chunk length
        0x74, 0x52, 0x4E, 0x53,                                     // tRNS, transparency index begin
        0x60,                                                       // Opacity (transparency)
        0x0D, 0x54, 0xB9, 0x3E,                                     // tRNS CRC
        0x00, 0x00, 0x00, 0x1F,                                     // IDAT length
        0x49, 0x44, 0x41, 0x54,                                     // IDAT chunk beginning
        0x78, 0xDA, 0xED, 0xC1, 0x31, 0x01, 0x00,                   // Z
        0x00, 0x00, 0xC2, 0xA0, 0xF5, 0x4F, 0x6D,                   // Z
        0x0D, 0x0F, 0xA0, 0x00, 0x00, 0x00, 0x00,                   // Z
        0x00, 0x00, 0x00, 0x00, 0xAE, 0x0D, 0x20,                   // Z
        0xDF, 0x00, 0x01,                                           // Z
        0xDB, 0x24, 0x3B, 0x5D,                                     // IDAT CRC
        0x00, 0x00, 0x00, 0x00,                                     // IEND chunk length
        0x49, 0x45, 0x4E, 0x44,                                     // IEND, eof signature 
        0xAE, 0x42, 0x60, 0x82                                      // IEND CRC
    };
    
    FMX_Hdl result = FMX_NewHandle( sizeof(solidPng) );             // Allocate memory for alpha PNG
    unsigned char palette[] = {0x50, 0x4C, 0x54, 0x45, R, G, B};    // Palette array
    long          palcrc = crc(palette, sizeof(palette));           // Generation of CRC from palette
    unsigned char opaq[] = {0x74, 0x52, 0x4E, 0x53, A};             // Transparency array
    long          opaqcrc = crc(opaq, sizeof(opaq));                // Generation of CRC from transparency

    if( result )
    {
        memcpy( *result, solidPng, sizeof(solidPng) );              // Copy default PNG to the buffer
        
        (*result)[0x29] = (unsigned char)R;                         // Modify the PNG's color palette
        (*result)[0x2A] = (unsigned char)G; 
        (*result)[0x2B] = (unsigned char)B;
        (*result)[0x2C] = (unsigned char)(palcrc >> 24);            // CRC of color palette
        (*result)[0x2D] = (unsigned char)(palcrc >> 16); 
        (*result)[0x2E] = (unsigned char)(palcrc >> 8);
        (*result)[0x2F] = (unsigned char)(palcrc);
        (*result)[0x38] = (unsigned char)A;                         // Modify transparency
        (*result)[0x39] = (unsigned char)(opaqcrc >> 24);           //  and CRC of transparency
        (*result)[0x3A] = (unsigned char)(opaqcrc >> 16);
        (*result)[0x3B] = (unsigned char)(opaqcrc >> 8);
        (*result)[0x3C] = (unsigned char)(opaqcrc); 
    }
    
    return result;
}

static FMX_Hdl CreateSinglePixelJPEG()

// It contains 1x1 blue pixel. Used with PICT and PNG.
// In final result, Filemaker ignores it and displays only PICT or PNG. 

{
    unsigned char solidJpg[] = { 
        0xFF, 0xD8, 0xFF, 0xE0, 0x00, 0x10, 0x4A, 
        0x46, 0x49, 0x46, 0x00, 0x01, 0x01, 0x01, 
        0x00, 0x47, 0x00, 0x47, 0x00, 0x00, 0xFF, 
        0xFE, 0x00, 0x0C, 0x41, 0x70, 0x70, 0x6C, 
        0x65, 0x4D, 0x61, 0x72, 0x6B, 0x0A, 0xFF, 
        0xDB, 0x00, 0x84, 0x00, 0x07, 0x05, 0x05, 
        0x06, 0x05, 0x05, 0x07, 0x06, 0x06, 0x06, 
        0x08, 0x07, 0x07, 0x08, 0x0A, 0x11, 0x0B, 
        0x0A, 0x09, 0x09, 0x0A, 0x14, 0x0F, 0x0F, 
        0x0C, 0x11, 0x18, 0x15, 0x19, 0x19, 0x17, 
        0x15, 0x17, 0x17, 0x1A, 0x1D, 0x25, 0x20, 
        0x1A, 0x1C, 0x23, 0x1C, 0x17, 0x17, 0x21, 
        0x2C, 0x21, 0x23, 0x27, 0x28, 0x2A, 0x2A, 
        0x2A, 0x19, 0x1F, 0x2E, 0x31, 0x2D, 0x29, 
        0x31, 0x25, 0x29, 0x2A, 0x28, 0x01, 0x07, 
        0x08, 0x08, 0x0A, 0x09, 0x0A, 0x13, 0x0B, 
        0x0B, 0x13, 0x28, 0x1B, 0x17, 0x1B, 0x28, 
        0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 
        0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 
        0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 
        0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 
        0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 
        0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 
        0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 0x28, 
        0xFF, 0xC4, 0x01, 0xA2, 0x00, 0x00, 0x01, 
        0x05, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
        0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 
        0x07, 0x08, 0x09, 0x0A, 0x0B, 0x01, 0x00, 
        0x03, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
        0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 
        0x00, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 
        0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x10, 
        0x00, 0x02, 0x01, 0x03, 0x03, 0x02, 0x04, 
        0x03, 0x05, 0x05, 0x04, 0x04, 0x00, 0x00, 
        0x01, 0x7D, 0x01, 0x02, 0x03, 0x00, 0x04, 
        0x11, 0x05, 0x12, 0x21, 0x31, 0x41, 0x06, 
        0x13, 0x51, 0x61, 0x07, 0x22, 0x71, 0x14, 
        0x32, 0x81, 0x91, 0xA1, 0x08, 0x23, 0x42, 
        0xB1, 0xC1, 0x15, 0x52, 0xD1, 0xF0, 0x24, 
        0x33, 0x62, 0x72, 0x82, 0x09, 0x0A, 0x16, 
        0x17, 0x18, 0x19, 0x1A, 0x25, 0x26, 0x27, 
        0x28, 0x29, 0x2A, 0x34, 0x35, 0x36, 0x37, 
        0x38, 0x39, 0x3A, 0x43, 0x44, 0x45, 0x46, 
        0x47, 0x48, 0x49, 0x4A, 0x53, 0x54, 0x55, 
        0x56, 0x57, 0x58, 0x59, 0x5A, 0x63, 0x64, 
        0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x73, 
        0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A, 
        0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 
        0x8A, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 
        0x98, 0x99, 0x9A, 0xA2, 0xA3, 0xA4, 0xA5, 
        0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xB2, 0xB3, 
        0xB4, 0xB5, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 
        0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7, 0xC8, 
        0xC9, 0xCA, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 
        0xD7, 0xD8, 0xD9, 0xDA, 0xE1, 0xE2, 0xE3, 
        0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 
        0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 
        0xF8, 0xF9, 0xFA, 0x11, 0x00, 0x02, 0x01, 
        0x02, 0x04, 0x04, 0x03, 0x04, 0x07, 0x05, 
        0x04, 0x04, 0x00, 0x01, 0x02, 0x77, 0x00, 
        0x01, 0x02, 0x03, 0x11, 0x04, 0x05, 0x21, 
        0x31, 0x06, 0x12, 0x41, 0x51, 0x07, 0x61, 
        0x71, 0x13, 0x22, 0x32, 0x81, 0x08, 0x14, 
        0x42, 0x91, 0xA1, 0xB1, 0xC1, 0x09, 0x23, 
        0x33, 0x52, 0xF0, 0x15, 0x62, 0x72, 0xD1, 
        0x0A, 0x16, 0x24, 0x34, 0xE1, 0x25, 0xF1, 
        0x17, 0x18, 0x19, 0x1A, 0x26, 0x27, 0x28, 
        0x29, 0x2A, 0x35, 0x36, 0x37, 0x38, 0x39, 
        0x3A, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 
        0x49, 0x4A, 0x53, 0x54, 0x55, 0x56, 0x57, 
        0x58, 0x59, 0x5A, 0x63, 0x64, 0x65, 0x66, 
        0x67, 0x68, 0x69, 0x6A, 0x73, 0x74, 0x75, 
        0x76, 0x77, 0x78, 0x79, 0x7A, 0x82, 0x83, 
        0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8A, 
        0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 
        0x99, 0x9A, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 
        0xA7, 0xA8, 0xA9, 0xAA, 0xB2, 0xB3, 0xB4, 
        0xB5, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xC2, 
        0xC3, 0xC4, 0xC5, 0xC6, 0xC7, 0xC8, 0xC9, 
        0xCA, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 0xD7, 
        0xD8, 0xD9, 0xDA, 0xE2, 0xE3, 0xE4, 0xE5, 
        0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xF2, 0xF3, 
        0xF4, 0xF5, 0xF6, 0xF7, 0xF8, 0xF9, 0xFA, 
        0xFF, 0xC0, 0x00, 0x11, 0x08, 0x00, 0x01, 
        0x00, 0x01, 0x03, 0x01, 0x22, 0x00, 0x02, 
        0x11, 0x01, 0x03, 0x11, 0x01, 0xFF, 0xDA, 
        0x00, 0x0C, 0x03, 0x01, 0x00, 0x02, 0x11, 
        0x03, 0x11, 0x00, 0x3F, 0x00, 0xE1, 0x28, 
        0xA2, 0x8A, 0xF8, 0xE3, 0xFA, 0x48, 0xFF, 
        0xD9 };
        
    FMX_Hdl result = FMX_NewHandle( sizeof(solidJpg) );             // Allocate memory for solid JPEG
 
    if( result )
    {
        memcpy( *result, solidJpg, sizeof(solidJpg) );              // Copy default solid JPEG to the buffer
    }
    
    return result;      
}
